import { CustomObject, LayoutCoordinator, LayoutCoordinatorExtra } from './Interfaces';
export declare class TreeStructure implements LayoutCoordinator {
    private customObjectSpace;
    private orientation;
    private groups;
    private sortedGroups;
    private customObjects;
    private coordinates;
    private distances;
    constructor({ customObjectSpace, orientation }: LayoutCoordinatorExtra);
    /**
     * Prepares and organizes objects for further processing by organizing them into groups and sorting the groups.
     * @param {CustomObject[]} customObjects - The objects to be prepared.
     */
    prepare: (customObjects: CustomObject[]) => void;
    /**
     * Executes the processing steps for calculating distances between objects.
     * This method triggers the calculation of distances between objects.
     */
    execute: () => void;
    /**
     * Finds and returns the parent object of a given object from a list of objects based on their parent-child relationships.
     * @param {CustomObjectFormated[]} customObjects - The list of objects to search for the parent.
     * @param {CustomObjectFormated} currentCustomObject - The object for which to find the parent.
     * @returns {CustomObjectFormated | undefined} The parent object, or undefined if no parent is found.
     */
    private getParentFromObjects;
    /**
     * Organizes objects into groups based on their parent-child relationships using a level-wise traversal.
     * @param {CustomObjectFormated[]} customObjects - The objects to be organized into groups.
     */
    private organizeObjectsIntoGroups;
    /**
     * Recursively sorts the children of a object based on the number of their children,
     * creating a hierarchical structure of sorted groups.
     *
     * @param {CustomObjectFormated[]} children - The children objects to be sorted.
     * @returns {SortedGroups[]} An array of sorted groups with root objects and their sorted children.
     */
    private sortChildren;
    /**
     * Sorts groups based on the length of their elements and organizes them with their root rectangles and sorted children.
     */
    private sortGroups;
    /**
     * Calculates the starting point of a smaller line within a larger container.
     * @param {number} left - The left position of the larger container.
     * @param {number} width - The width of the smaller line.
     * @returns {number} The starting point X-coordinate of the smaller line.
     */
    private calculateStartingPointOfSmallerLine;
    /**
     * Calculates the distance based on the orientation.
     * If the orientation is vertical, it calls distanceVertical();
     * otherwise, it calls distanceHorizontal().
     */
    private distance;
    /**
     * Distributes children along the specified axis within a parent container.
     * @param children - An array of sorted groups representing child elements.
     * @param axis - The axis along which the children should be distributed.
     * @param parentSize - The size of the parent container along the distribution axis.
     * @param parentEnd - The end position of the parent container along the distribution axis.
     */
    private distributeChildren;
    /**
     * Handles the horizontal orientation layout for a object and its children.
     * @param customObject - The sorted groups representing the object and its children.
     * @param x - The initial x-coordinate for positioning.
     * @param y - The y-coordinate for positioning.
     * @param childTopy - The top y-coordinate of child elements.
     * @param leftx - The left x-coordinate of the object.
     * @returns The updated x and y coordinates.
     */
    private handleHorizontalOrientation;
    /**
     * Handles the vertical orientation layout for a object and its children.
     * @param customObject - The sorted groups representing the object and its children.
     * @param x - The x-coordinate for positioning.
     * @param y - The initial y-coordinate for positioning.
     * @param topy - The top y-coordinate of the object.
     * @param childLeftx - The left x-coordinate of child elements.
     * @returns The updated x and y coordinates.
     */
    private handleVerticalOrientation;
    /**
     * Sets the position based on orientation and index for a object.
     * @param customObject - The sorted groups representing the object.
     * @param x - The initial x-coordinate for positioning.
     * @param y - The initial y-coordinate for positioning.
     * @param leftx - The left x-coordinate of the object.
     * @param topy - The top y-coordinate of the object.
     * @param orientation - The orientation of the layout (horizontal or vertical).
     * @param index - The index of the object.
     * @returns The updated x and y coordinates.
     */
    private setPositionBasedOnOrientationAndIndex;
    /**
     * Calculates the position based on orientation and child elements for a object.
     * @param child - The sorted groups representing child elements.
     * @param topy - The top y-coordinate of the parent object.
     * @param leftx - The left x-coordinate of the parent object.
     * @param orientation - The orientation of the layout (horizontal or vertical).
     * @returns The calculated position.
     */
    private childDistances;
    /**
     * Calculates the distance between groups and rectangles, positioning them based on their relationships in a horizontal manner.
     *
     * @description This method organizes the groups and their children, calculating positions and distances
     * to create a visually balanced layout.
     */
    private distanceHorizontal;
    /**
     * Calculates the distance between groups and rectangles, positioning them based on their relationships in a vertical manner.
     *
     * @description This method organizes the groups and their children, calculating positions and distances
     * to create a visually balanced layout.
     */
    private distanceVertical;
}
